@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DataKinds #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE GADTs #-}
@@ -19,6 +20,7 @@ module Database.Persist.Postgresql
     , openSimpleConn
     , tableName
     , fieldName
+    , mockMigration
     ) where
 
 import Database.Persist.Sql
@@ -39,6 +41,7 @@ import qualified Database.PostgreSQL.LibPQ as LibPQ
 import Control.Monad.Trans.Resource
 import Control.Exception (throw)
 import Control.Monad.IO.Class (MonadIO (..))
+import Data.Data
 import Data.Typeable
 import Data.IORef
 import qualified Data.Map as Map
@@ -57,12 +60,15 @@ import Data.ByteString (ByteString)
 import qualified Data.ByteString.Char8 as B8
 import qualified Data.Text as T
 import qualified Data.Text.Encoding as T
+import qualified Data.Text.IO as T
 import qualified Blaze.ByteString.Builder.Char8 as BBB
 
 import Data.Text (Text)
 import Data.Aeson
 import Data.Aeson.Types (modifyFailure)
 import Control.Monad (forM)
+import Control.Monad.Trans.Reader (runReaderT)
+import Control.Monad.Trans.Writer (runWriterT)
 import Data.Acquire (Acquire, mkAcquire, with)
 import System.Environment (getEnvironment)
 import Data.Int (Int64)
@@ -147,6 +153,7 @@ openSimpleConn logFunc conn = do
         { connPrepare    = prepare' conn
         , connStmtMap    = smap
         , connInsertSql  = insertSql'
+        , connInsertManySql = Just insertManySql'
         , connClose      = PG.close conn
         , connMigrateSql = migrate'
         , connBegin      = const $ PG.begin    conn
@@ -188,6 +195,21 @@ insertSql' ent vals =
        Just _pdef -> ISRManyKeys sql vals
        Nothing -> ISRSingle (sql <> " RETURNING " <> escape (fieldDB (entityId ent)))
 
+-- | SQL for inserting multiple rows at once and returning their primary keys.
+insertManySql' :: EntityDef -> [[PersistValue]] -> InsertSqlResult
+insertManySql' ent valss =
+  let sql = T.concat
+                [ "INSERT INTO "
+                , escape (entityDB ent)
+                , "("
+                , T.intercalate "," $ map (escape . fieldDB) $ entityFields ent
+                , ") VALUES ("
+                , T.intercalate "),(" $ replicate (length valss) $ T.intercalate "," $ map (const "?") (entityFields ent)
+                , ") RETURNING "
+                , T.intercalate ", " $ dbIdColumnsEsc escape ent
+                ]
+  in ISRSingle sql
+
 execute' :: PG.Connection -> PG.Query -> [PersistValue] -> IO Int64
 execute' conn query vals = PG.execute conn query (map P vals)
 
@@ -246,7 +268,12 @@ withStmt' conn query vals =
            else fmap Just $ forM (zip getters [0..]) $ \(getter, col) -> do
                                 mbs <- LibPQ.getvalue' ret row col
                                 case mbs of
-                                  Nothing -> return PersistNull
+                                  Nothing ->
+                                    -- getvalue' verified that the value is NULL.
+                                    -- However, that does not mean that there are
+                                    -- no NULL values inside the value (e.g., if
+                                    -- we're dealing with an array of optional values).
+                                    return PersistNull
                                   Just bs -> do
                                     ok <- PGFF.runConversion (getter mbs) conn
                                     bs `seq` case ok of
@@ -283,7 +310,7 @@ newtype Unknown = Unknown { unUnknown :: ByteString }
 instance PGFF.FromField Unknown where
     fromField f mdata =
       case mdata of
-        Nothing  -> PGFF.returnError PGFF.UnexpectedNull f ""
+        Nothing  -> PGFF.returnError PGFF.UnexpectedNull f "Database.Persist.Postgresql/PGFF.FromField Unknown"
         Just dat -> return (Unknown dat)
 
 instance PGTF.ToField Unknown where
@@ -307,8 +334,10 @@ builtinGetters = I.fromList
     , (k PS.xml,         convertPV PersistText)
     , (k PS.float4,      convertPV PersistDouble)
     , (k PS.float8,      convertPV PersistDouble)
+#if !MIN_VERSION_postgresql_simple(0,5,0)
     , (k PS.abstime,     convertPV PersistUTCTime)
     , (k PS.reltime,     convertPV PersistUTCTime)
+#endif
     , (k PS.money,       convertPV PersistRational)
     , (k PS.bpchar,      convertPV PersistText)
     , (k PS.varchar,     convertPV PersistText)
@@ -324,9 +353,8 @@ builtinGetters = I.fromList
     , (k PS.jsonb,       convertPV (PersistByteString . unUnknown))
     , (k PS.unknown,     convertPV (PersistByteString . unUnknown))
 
-
-
-    -- array types: same order as above
+    -- Array types: same order as above.
+    -- The OIDs were taken from pg_type.
     , (1000,             listOf PersistBool)
     , (1001,             listOf (PersistByteString . unBinary))
     , (1002,             listOf PersistText)
@@ -358,7 +386,13 @@ builtinGetters = I.fromList
     ]
     where
         k (PGFF.typoid -> i) = PG.oid2int i
-        listOf f = convertPV (PersistList . map f . PG.fromPGArray)
+        -- A @listOf f@ will use a @PGArray (Maybe T)@ to convert
+        -- the values to Haskell-land.  The @Maybe@ is important
+        -- because the usual way of checking NULLs
+        -- (c.f. withStmt') won't check for NULL inside
+        -- arrays---or any other compound structure for that matter.
+        listOf f = convertPV (PersistList . map (nullable f) . PG.fromPGArray)
+          where nullable = maybe PersistNull
 
 getGetter :: PG.Connection -> PG.Oid -> Getter PersistValue
 getGetter _conn oid
@@ -375,7 +409,8 @@ doesTableExist getter (DBName name) = do
     stmt <- getter sql
     with (stmtQuery stmt vals) ($$ start)
   where
-    sql = "SELECT COUNT(*) FROM information_schema.tables WHERE table_name=?"
+    sql = "SELECT COUNT(*) FROM pg_catalog.pg_tables WHERE schemaname != 'pg_catalog'"
+          <> " AND schemaname != 'information_schema' AND tablename=?"
     vals = [PersistText name]
 
     start = await >>= maybe (error "No results when checking doesTableExist") start'
@@ -416,18 +451,8 @@ migrate' allDefs getter entity = fmap (fmap $ map showAlterDb) $ do
             -- for https://github.com/yesodweb/persistent/issues/152
 
     createText newcols fdefs udspair =
-        addTable : uniques ++ references ++ foreignsAlt
+        (addTable newcols entity) : uniques ++ references ++ foreignsAlt
       where
-        addTable = AddTable $ T.concat
-                -- Lower case e: see Database.Persist.Sql.Migration
-                [ "CREATe TABLE " -- DO NOT FIX THE CAPITALIZATION!
-                , escape name
-                , "("
-                , idtxt
-                , if null newcols then "" else ","
-                , T.intercalate "," $ map showColumn newcols
-                , ")"
-                ]
         uniques = flip concatMap udspair $ \(uname, ucols) ->
                 [AlterTable name $ AddUniqueConstraint uname ucols]
         references = mapMaybe (\c@Column { cName=cname, cReference=Just (refTblName, _) } ->
@@ -437,7 +462,20 @@ migrate' allDefs getter entity = fmap (fmap $ map showAlterDb) $ do
             let (childfields, parentfields) = unzip (map (\((_,b),(_,d)) -> (b,d)) (foreignFields fdef))
             in AlterColumn name (foreignRefTableDBName fdef, AddReference (foreignConstraintNameDBName fdef) childfields (map escape parentfields)))
 
-    idtxt = case entityPrimary entity of
+addTable :: [Column] -> EntityDef -> AlterDB
+addTable cols entity = AddTable $ T.concat
+                       -- Lower case e: see Database.Persist.Sql.Migration
+                       [ "CREATe TABLE " -- DO NOT FIX THE CAPITALIZATION!
+                       , escape name
+                       , "("
+                       , idtxt
+                       , if null cols then "" else ","
+                       , T.intercalate "," $ map showColumn cols
+                       , ")"
+                       ]
+    where
+      name = entityDB entity
+      idtxt = case entityPrimary entity of
                 Just pdef -> T.concat [" PRIMARY KEY (", T.intercalate "," $ map (escape . fieldDB) $ compositeFields pdef, ")"]
                 Nothing   ->
                     let defText = defaultAttribute $ fieldAttrs $ entityId entity
@@ -449,7 +487,6 @@ migrate' allDefs getter entity = fmap (fmap $ map showAlterDb) $ do
                             , mayDefault defText
                             ]
 
-
 maySerial :: SqlType -> Maybe Text -> Text
 maySerial SqlInt64 Nothing = " SERIAL "
 maySerial sType _ = " " <> showSqlType sType
@@ -892,10 +929,10 @@ data PostgresConf = PostgresConf
       -- ^ The connection string.
     , pgPoolSize :: Int
       -- ^ How many connections should be held on the connection pool.
-    } deriving Show
+    } deriving (Show, Read, Data, Typeable)
 
 instance FromJSON PostgresConf where
-    parseJSON v = modifyFailure ("Persistent: error loadomg PostgreSQL conf: " ++) $
+    parseJSON v = modifyFailure ("Persistent: error loading PostgreSQL conf: " ++) $
       flip (withObject "PostgresConf") v $ \o -> do
         database <- o .: "database"
         host     <- o .: "host"
@@ -953,3 +990,70 @@ refName (DBName table) (DBName column) =
 
 udToPair :: UniqueDef -> (DBName, [DBName])
 udToPair ud = (uniqueDBName ud, map snd $ uniqueFields ud)
+
+mockMigrate :: [EntityDef]
+         -> (Text -> IO Statement)
+         -> EntityDef
+         -> IO (Either [Text] [(Bool, Text)])
+mockMigrate allDefs _ entity = fmap (fmap $ map showAlterDb) $ do
+    case partitionEithers [] of
+        ([], old'') -> return $ Right $ migrationText False old''
+        (errs, _) -> return $ Left errs
+  where
+    name = entityDB entity
+    migrationText exists old'' =
+        if not exists
+            then createText newcols fdefs udspair
+            else let (acs, ats) = getAlters allDefs entity (newcols, udspair) old'
+                     acs' = map (AlterColumn name) acs
+                     ats' = map (AlterTable name) ats
+                 in  acs' ++ ats'
+       where
+         old' = partitionEithers old''
+         (newcols', udefs, fdefs) = mkColumns allDefs entity
+         newcols = filter (not . safeToRemove entity . cName) newcols'
+         udspair = map udToPair udefs
+            -- Check for table existence if there are no columns, workaround
+            -- for https://github.com/yesodweb/persistent/issues/152
+
+    createText newcols fdefs udspair =
+        (addTable newcols entity) : uniques ++ references ++ foreignsAlt
+      where
+        uniques = flip concatMap udspair $ \(uname, ucols) ->
+                [AlterTable name $ AddUniqueConstraint uname ucols]
+        references = mapMaybe (\c@Column { cName=cname, cReference=Just (refTblName, _) } ->
+            getAddReference allDefs name refTblName cname (cReference c))
+                   $ filter (isJust . cReference) newcols
+        foreignsAlt = flip map fdefs (\fdef ->
+            let (childfields, parentfields) = unzip (map (\((_,b),(_,d)) -> (b,d)) (foreignFields fdef))
+            in AlterColumn name (foreignRefTableDBName fdef, AddReference (foreignConstraintNameDBName fdef) childfields (map escape parentfields)))
+
+-- | Mock a migration even when the database is not present.
+-- This function performs the same functionality of 'printMigration'
+-- with the difference that an actualy database isn't needed for it.
+mockMigration :: Migration -> IO ()
+mockMigration mig = do
+  smap <- newIORef $ Map.empty
+  let sqlbackend = SqlBackend { connPrepare = \_ -> do
+                                             return Statement
+                                                        { stmtFinalize = return ()
+                                                        , stmtReset = return ()
+                                                        , stmtExecute = undefined
+                                                        , stmtQuery = \_ -> return $ return ()
+                                                        },
+                             connInsertManySql = Nothing,
+                             connInsertSql = undefined,
+                             connStmtMap = smap,
+                             connClose = undefined,
+                             connMigrateSql = mockMigrate,
+                             connBegin = undefined,
+                             connCommit = undefined,
+                             connRollback = undefined,
+                             connEscapeName = escape,
+                             connNoLimit = undefined,
+                             connRDBMS = undefined,
+                             connLimitOffset = undefined,
+                             connLogFunc = undefined}
+      result = runReaderT $ runWriterT $ runWriterT mig
+  resp <- result sqlbackend
+  mapM_ T.putStrLn $ map snd $ snd resp
